#!/usr/local/bin/perl5
#
# ABSOLUTELY NO WARRANTY WITH THIS PACKAGE. USE IT AT YOUR OWN RISK.
#
# Parse DIABLO logfile and generate stats about incoming & outgoing feeds.
# URL: ftp://ftp.ecrc.de/pub/news/servers/diablo-utils
#
# diablo-stats.pl v3.2  970816  Iain Lea  iain@ecrc.de
#
# Usage: diablo-stats.pl [options]
#     -h       help
#     -d date  date to gather statistics in YYMMDD (default: 970703)
#     -D days  number of days of statistics to display (default: all)
#     -B date  begin time to start gathering statistics in YYMMDD
#     -E date  end time to stop gathering statistics in YYMMDD
#     -g       generate various graphs of traffic statistics
#     -l file  diablo log file (default: /var/log/news/news.notice)
#     -o file  output file (default: stdout)
#     -s host  server hostname to use in report (default: newsfeed.ecrc.net)
#     -q       report queue by hostname instead of newsfeed name
#     -S       read logfile in pipe mode from STDIN
#     -t       output format as text (default: HTML)
#     -v       verbose
#     -V       debug
#     -w dir   create dir/index.html page to list daily statistics
#     -W dir   create dir/index.html page to list daily statistics and exit
#
# TODO:
# -----
# - add 'alias: real-hostname alias-to-use' to conf file
# - add header & footer boilerplate placeholders that are read in from file
# - FIX %3.2f formatting in ascii mode
# - FIX 'Goto top of table, Goto top of page'
# - add Kbit/sec for each feed & then delete the hh:mm
#   use the secs= value 
# - add a cmd line switch to delete files that are more than -D N days in index.html
# - add graphs to the report. (innreport.pl/GraphMaher?) Top 10 in/out line
#   - add a trend graph to the index.html for the last N days of in & out
#   - use Chart.pm
#
# Acknowledgements:
# -----------------
#   Jeff Garzik             jeff.garzik@spinne.com
#   Terry Kennedy           terry@spcvxa.spc.edu
#   Steve Rawlinson         steve@clara.net
#   Pierre Belanger         belanger@risq.qc.ca
#   Miquel van Smoorenburg  miquels@cistron.nl
#
# ChangeLog:
# ----------
# v3.2
#   - added by-hour table showing total arts & vol per hour
#   - added http links to feedname stats page for reverse feed checks:
#     link:  news.foo.com  http://news.foo.com/stats/
#     link:  feed.bar.com  http://feef.bar.com/diablo/
#     etc. etc. (default file is /news/diablo-stats.conf)
#   - added SPAM totals for incoming feeds to top level index.html file
#   - added logo to page headers (default is ecrclogo.gif) + link to www site
#   - added $BgColor to <body> tags for defining background color 
#   - added rank field to incoming & outgoing tables
#   - added check for 'lost backchannel to master server' in logfile
#   - added 1stDate-LastDate to curious table instead of just 1stDate
#   - changed fields in incoming & outgoing tables for readability
#   - changed index.html to place Total+Volume as first fields
#   - changed the 'Total' field text to be centered
#   - fixed another BigFloat problem when no initialized value is used
#   - fixed BigFloat panic line 31 error by initializing $InBytes to 0
#   - fixed sorted order of >4GB feeds by using &MakeVolKey
# v3.0 970801
#   - added SPAM table 
#   - added more checks for '400/500/502' type error messages in logfile

use Math::BigInt;
use Math::BigFloat;
require 'getopts.pl';
require "timelocal.pl";

$CfgFile = "/news/diablo-stats.conf";
$LogFile = "/var/log/news/news.notice";
$CtlFile = "/news/dnntpspool.ctl";
$QueueCmd = "/news/dbin/doutq";
$ImgDir = "";
$ImgLogo = "$ImgDir/ecrclogo.gif";
$FtpUrl = "ftp://ftp.ecrc.de/pub/news/servers/diablo-utils/";
$WwwUrl = "http://www.ecrc.de/";
$ImgText = "www.ecrc.de";
$ImgSize = "height=\"50\" width=\"50\"";
# $ImgSize = "height=\"50\" width=\"50\" hspace=\"5\" vspace=\"5\";
$BgColor = "bgcolor=\"#FFFFFF\"";
$OutFile = "";
$HtmlDir = "";
$Verbose = 0;
$Debug = 0;
$Version = 'v3.2';
$MaxSpamList = 25;

$DiabloUrl = "http://www.backplane.com/diablo/";
$QueueHost = 0;
$TextFormat = 0;
$GraphMode = 0;
$TotalSpamList = 0;
$ScriptName = 'diablo-stats';
chop ($DiabloHost = `hostname`);
chop ($GenYear = `date +%y`);
chop ($GenDate = `date +%y%m%d`);
chop ($GenTime = `date +%H%M`);
$MaxDays = 0;
$CreateIndexAndExit = 0;
$TotalRequests = 0;
@HalfHours = 0;
$BegDate = $EndDate = $GenDate;
$BegTime = $EndTime = "";
$TDL = "<td align=\"left\">";
$TDR = "<td align=\"right\">";
@DayName = ("Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday");
@DayNameShort = ("Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat");
%MonthNumByName = (
	'Jan', 1, 'Feb', 2, 'Mar', 3, 'Apr', 4,  'May', 5,  'Jun', 6,
	'Jul', 7, 'Aug', 8, 'Sep', 9, 'Oct', 10, 'Nov', 11, 'Dec', 12);
%MonthNameByNum = (
	'01', 'Jan', '02', 'Feb', '03', 'Mar', '04', 'Apr', '05', 'May', '06', 'Jun',
	'07', 'Jul', '08', 'Aug', '09', 'Sep', '10', 'Oct', '11', 'Nov', '12', 'Dec');
$BgColor="bgcolor=\"#FFFFFF\"";
# Colorize outgoing queue. Defaults are 20/25/50/65/80/95%
$QueColor1 = 'lightgreen';
$QueColor2 = 'green';
$QueColor3 = 'lightblue';
$QueColor4 = 'lightyellow';
$QueColor5 = 'orange';
$QueColor6 = 'red';
$QuePercent1 = 20;
$QuePercent2 = 35;
$QuePercent3 = 50;
$QuePercent4 = 65;
$QuePercent5 = 80;
$QuePercent6 = 95;

##############################################################################
# 
#

&ParseCmdLine ($0);

# &InitReqsPerHalfHour;

if ($CreateIndexAndExit) {
	&CreateHtmlIndex;
} else {
	&ReadCfgFile;

	&ParseLogFile;

	&ParseOutQueue;

	&GenerateOutputFile;

	&CreateHtmlIndex;
}

exit 0;

##############################################################################
# 
#

sub ParseCmdLine
{
	my ($ProgName) = @_;
	
	&Getopts('B:E:d:D:ghl:o:s:qStvVw:W:');

	if ($opt_h) { 
		print <<EOT
$ScriptName $Version  $FtpUrl

Create statistics for DIABLO news relay server in Ascii / HTML format.
Copyright 1997 Iain Lea (iain\@ecrc.de).    NOTE: Use at your own risk.

Usage: $ProgName [options]
       -h       help
       -d date  date to gather statistics in YYMMDD (default: $GenDate)
       -D days  number of days of statistics to display (default: all)
       -B date  begin time to start gathering statistics in YYMMDD
       -E date  end time to stop gathering statistics in YYMMDD
       -g       generate various graphs of traffic statistics
       -l file  diablo log file (default: $LogFile)
       -o file  output file (default: stdout)
       -s host  server hostname to use in report (default: $DiabloHost)
       -q       report queue by hostname instead of newsfeed name
       -S       read logfile in pipe mode from STDIN
       -t       output format as text (default: HTML)
       -v       verbose
       -V       debug
       -w dir   create dir/index.html page to list daily statistics
       -W dir   create dir/index.html page to list daily statistics and exit

EOT
;
		exit 1;
	}
	$BegDate = $opt_B if (defined($opt_B));
	$EndDate = $opt_E if (defined($opt_E));
	$BegDate = $EndDate = $GenDate = $opt_d if (defined($opt_d));
	$MaxDays = $opt_D if (defined($opt_D));
	$GraphMode++ if (defined($opt_g));
	$LogFile = $opt_l if (defined($opt_l));
	$OutFile = $opt_o if (defined($opt_o));
	$DiabloHost = $opt_s if (defined($opt_s));
	$QueueHost++ if (defined($opt_q));
	$TextFormat++ if (defined($opt_t));
	$Verbose++ if (defined($opt_v));
	$Debug++ if (defined($opt_V));
	$HtmlDir = $opt_w if (defined($opt_w));

	if (defined($opt_S)) {
		$LogFile = "-";
	}

	if (defined($opt_W)) {
		$HtmlDir = $opt_W;
		$CreateIndexAndExit++;
	} else {
		die "Error: $LogFile - $!\n" if ($LogFile eq '');
	}
}


sub ParseLogFile
{
	my ($Error, $Line, $FullDate, $LineDate);

	printf STDERR "Parsing %s for $BegDate - $EndDate ...\n",
		 ($LogFile eq "-" ? "STDIN" : $LogFile) if $Verbose;

	open (FILE, $LogFile) || die "Error: $LogFile - $!\n";

	while ($Line = <FILE>) 
	{
		next unless (($Line =~ /diablo/o) || ($Line =~ /newslink/o));

		if ($Line =~ /^(... .. ..:..:..)/o) {
			
			$FullDate = $1;
			$LineDate = &GetYyMmDd ($FullDate);
			if ($LineDate >= $BegDate && $LineDate <= $EndDate) {
				chop $Line;

				# ... diablo[423]: secs=3120 ihave=0 chk=1536 rec=1024 rej=156 predup=25 posdup=0 pcoll=10 spam=0 err=0 added=868 (0/sec)
				if ($Line =~ /(.*)\s+diablo\[(\d+)\]:\s+(.*)\s+secs=(\d+)\s+ihave=(\d+)\s+chk=(\d+)\s+rec=(\d+)\s+rej=(\d+)\s+.*\s+err=(\d+)\s+added=(\d+)/o) {
#					$DiabloHost = $1;
#					$DiabloPid = $2;
					$InHost = $3;
					$InSecs = $4;
					$InIhave = $5;
					$InChk = $6;
					$InRec = $7;
					$InRej = $8;
					$InErr = $9;
					$InAdded = $10;

					# Diablo >= 1.10  ... predup=25 posdup=0 pcoll=10 spam=0 ... bytes=18908
					if ($Line =~ /predup=(\d+)\s+posdup=(\d+)\s+pcoll=(\d+)\s+spam=(\d+).*bytes=(\d+)/o) {
						$InPredup = $1;
						$InPosdup = $2;
						$InPcoll = $3;
						$InSpam = $4;
						$InBytes = $5;
					}

					$InHost =~ s/\s//g;
					$BegTime = $LineDate if !$BegTime;
					$EndTime = $LineDate;

					$Feeds{$InHost}->{InCons} += 1;
					$Feeds{$InHost}->{InSecs} += $InSecs;
					$Feeds{$InHost}->{InIhave} += $InIhave;
					$Feeds{$InHost}->{InChk} += $InChk;
					$Feeds{$InHost}->{InRec} += $InRec;
					$Feeds{$InHost}->{InRej} += $InRej;
					$Feeds{$InHost}->{InPredup} += $InPredup;
					$Feeds{$InHost}->{InPosdup} += $InPosdup;
					$Feeds{$InHost}->{InPcoll} += $InPcoll;
					$Feeds{$InHost}->{InSpam} += $InSpam;
					$Feeds{$InHost}->{InErr} += $InErr;
					$Feeds{$InHost}->{InAdded} += $InAdded;
					$Feeds{$InHost}->{InBytes} += $InBytes;

					&AddToInTimeList ($FullDate, $InHost, $InAdded, $InBytes, 
						$InChk, $InIhave, $InSpam, $InRej, $InErr); # if $InAdded;

					printf STDERR "IN  [%s] %4d %4d %4d %4d %4d %4d %4d %4d %4d\n", 
						$InHost, $InSecs, $InIhave, $InChk, $InRec, $InRej, 
						$InSpam, $InErr, $InAdded, $InBytes if $Debug;

				# ... newslink[14596]: news.ecrc.de:news.ecrc.de.S00013 final secs=74 acc=0 dup=48 rej=1086 tot=1134 ...
				# ... newslink[22406]: news.ecrc.de:/news/dqueue/news.ecrc.de final secs=129 acc=4 dup=299 rej=0 tot=303 ...
				} elsif ($Line =~ /newslink\[(\d+)\]:\s+(.*):(.*)\s+final\s+secs=(\d+)\s+acc=(\d+)\s+dup=(\d+)\s+rej=(\d+)\s+tot=(\d+)/o) {

#					$NewslinkPid = $1;
					$OutHost = $2;
					$OutBatch = $3;
					$OutSecs = $4;
					$OutAcc = $5;
					$OutDup = $6;
					$OutRej = $7;

					$OutHost =~ s/\s//g;
					$BegTime = $LineDate if !$BegTime;
					$EndTime = $LineDate;

					if ($Line =~ /bytes=(\d+)(.*)/o) {
						$OutBytes = $1;
					}

					$Feeds{$OutHost}->{OutCons} += 1;
					$Feeds{$OutHost}->{OutSecs} += $OutSecs;
					$Feeds{$OutHost}->{OutAcc} += $OutAcc;
					$Feeds{$OutHost}->{OutDup} += $OutDup;
					$Feeds{$OutHost}->{OutRej} += $OutRej;
					$Feeds{$OutHost}->{OutBytes} += $OutBytes;

					&AddToOutTimeList ($FullDate, $OutHost, $OutAcc, 
						$OutBytes, $OutDup, $OutRej); # if $OutAcc;

					printf STDERR "OUT [%s] %4d %4d %4d %4d %4d %4d\n", 
						$OutHost, $OutSecs, $OutAcc, $OutDup, $OutRej, $OutBytes if $Debug;

				# ... newslink[7390]: news.crg.net:/news/dqueue/news.crg.net.S03454 connect: 500 Syntax error or bad command (nostreaming) ...
				} elsif ($Line =~ /newslink\[(\d+)\]:\s+(.*):(.*)\s+connect:\s+(.*)/o) {

#					$NewslinkPid = $1;
					$OutHost = $2;
					$OutBatch = $3;
					$OutMsg = $4;

					if ($OutMsg =~ /((400|500|502)\s+.*)/ ||
						$OutMsg =~ /(Connection timed out.*)/ ||
						$OutMsg =~ /(Connection refused.*)/) {
						$OutHost =~ s/\s//g;
						$Error = "$OutHost: $1";
						&AddToErrorList ($Error, $FullDate, $LineDate);
					}

				# ... diablo[423]: SpamFilter/by-post-rate copy #33: <5rplh4$12p@ron.ipa.com>  206.1.57.29
				# ... diablo[423]: SpamFilter/by-dup-body copy #1: <5rpvin$8ro$1@demdwu11.telemedia.de> essn-m103-83.pool.mediaways.net
				# ... diablo[423]: SpamFilter/dnewsfeeds copy #-1: <5r4dpg$le8$1@news.utu.fi> apus.astro.utu.fi
				} elsif ($Line =~ /diablo.*SpamFilter\/(.*)/o) {

					&AddToSpamList ($1, $FullDate, $LineDate);

				# ... diablo[423]: Connection 8 from 194.59.191.4 (no permission)
				} elsif ($Line =~ /diablo.*Connection.*from\s+(.*)\s+\(no permission\)/o) {

					$Error = "$1 (no permission)";
					&AddToErrorList ($Error, $FullDate, $LineDate);

				# ... diablo[423]: Connect Limit exceeded for 194.108.168.3
				} elsif ($Line =~ /diablo.*Connect Limit exceeded for\s+(.*)/o) {

					$Error = "Connect Limit exceeded for $1";
					&AddToErrorList ($Error, $FullDate, $LineDate);

				# ... diablo[423]: dhistory file corrupted on lookup
				# ... diablo[423]: dhistory file realigned by 4 @20439040
				} elsif ($Line =~ /diablo.*(dhistory file.*)/o) {

					&AddToErrorList ($1, $FullDate, $LineDate);

				# ... diablo[423]: lost backchannel to master server
				} elsif ($Line =~ /diablo.*(lost backchannel to master server)/o) {

					&AddToErrorList ($1, $FullDate, $LineDate);
				}
			}
		}
	}
	close (FILE);

	$Num = 0;
	foreach $Host (sort keys %Feeds) {
		# InAdded
		$Key = sprintf "%012lu.%03d", $Feeds{$Host}->{InAdded}, $Num;
		$OrderByInAdded{$Key} = $Host;
		# InBytes
		$Key = &MakeVolKey ($Feeds{$Host}->{InBytes}, $Num);
		$OrderByInBytes{$Key} = $Host;

		# OutAcc
		$Key = sprintf "%012lu.%03d", $Feeds{$Host}->{OutAcc}, $Num;
		$OrderByOutAcc{$Key} = $Host;
		# OutBytes
		$Key = &MakeVolKey ($Feeds{$Host}->{OutBytes}, $Num);
		$OrderByOutBytes{$Key} = $Host;

		$Num++;
	}

	$Num = 0;
	foreach $Msg (sort keys %ErrorList) {
		$Key = sprintf "%08lu.%03d%s", $ErrorList{$Msg}->{Tot}, $Num, $Msg;
		$OrderByErrorTotal{$Key} = $Msg;

		$Num++;
	}

	$Num = 0;
	foreach $Host (sort keys %SpamList) {
		# SpamTotal
		$Key = sprintf "%08lu.%03d", $SpamList{$Host}->{Total}, $Num;
		$OrderBySpamTotal{$Key} = $Host;

		$Num++;
	}
	$TotalSpamList = $Num;

	&DumpTimeList if $Debug;
}


sub ParseOutQueue
{
	my ($Key, $Host, $Num);

	print STDERR "Parsing output of $QueueCmd ...\n" if $Verbose;

	open (CMD, "$QueueCmd |") || die "Error: $QueueCmd - $!\n";

	while (<CMD>) 
	{
#		print if $Debug;

		chop;

		# news.ecrc.de  2699-2700  (  1/200 files   0% full)	02699 
		if (/^(.*)\s+(\d+)-(\d+)\s+\((.*)\/(.*)\s+files\s+(\d+)%\s+full\)/o) {

			$Host = $1;
			$BatchBeg = $2;
			$BatchEnd = $3;
			$BatchNum = $4;
			$BatchMax = $5;
			$BatchFull = $6;
			$Host =~ s/\s//g;

			$Feeds{$Host}->{BatchBeg} += $BatchBeg;
			$Feeds{$Host}->{BatchEnd} += $BatchEnd;
			$Feeds{$Host}->{BatchNum} += $BatchNum;
			$Feeds{$Host}->{BatchMax} += $BatchMax;
			$Feeds{$Host}->{BatchFull} += $BatchFull;

			printf STDERR "BATCH  %-30s %4d %4d %4d %4d %4d\n", 
				$Host, $Feeds{$Host}->{BatchBeg}, $Feeds{$Host}->{BatchEnd},
				$Feeds{$Host}->{BatchNum}, $Feeds{$Host}->{BatchMax},
				$Feeds{$Host}->{BatchFull} if $Debug;
		}
	}
	close CMD;

	$Num = 0;
	foreach $Host (sort keys %Feeds) {
		$Key = sprintf "%10d.%3d", $Feeds{$Host}->{BatchFull}, $Num;
		$OrderByBatchFull{$Key} = $Host;
		$Num++;
	}
}


sub GenerateOutputFile
{
#	print STDERR "OUT=[$OutFile]\n";

	if ($OutFile) {
		open (FILE, "> $OutFile") || warn "Warning: $OutFile - $!\n";
	} else {
		open (FILE, "> STDOUT") || warn "Warning: STDOUT - $!\n";
	}

	&PrintOutHeader;
 
	print FILE "<p>\n<a name=\"01\">\n<hr>\n<p>\n" if !$TextFormat;
	&PrintIncomingFeeds ("1. Summary of Incoming Feeds by Article", 0);

	&PrintSeparator ("02") if !$TextFormat;
	&PrintIncomingFeeds ("2. Summary of Incoming Feeds by Volume", 1);

	&PrintSeparator ("03") if !$TextFormat;
	&PrintIncomingTimes ("3. Summary of Incoming Feeds by Time");

#	&PrintChart (undef, "Incoming feeds by articles per half hour", undef);

	&PrintSeparator ("04") if !$TextFormat;
	&PrintOutgoingFeeds ("4. Summary of Outgoing Feeds by Article", 0);

	&PrintSeparator ("05") if !$TextFormat;
	&PrintOutgoingFeeds ("5. Summary of Outgoing Feeds by Volume", 1);

	&PrintSeparator ("06") if !$TextFormat;
	&PrintQueueStats ("6. Summary of Outgoing Queue");

	&PrintSeparator ("07") if !$TextFormat;
	&PrintCuriousActivity ("7. Summary of Curious Activity");

	&PrintSeparator ("08") if !$TextFormat;
	&PrintSpamActivity ("8. Summary of SPAM Activity (Top $MaxSpamList of $TotalSpamList)");

	&PrintSeparator ("end") if !$TextFormat;

	&PrintOutFooter;

	close FILE;
}


sub CreateHtmlIndex
{
	my ($IndexFile) = "$HtmlDir/index.html";
	my ($Title1) = "DIABLO news statistics for '$DiabloHost'";
	my ($Title2) = "<a href=$DiabloUrl>DIABLO</a>  news statistics for '$DiabloHost'";
	my ($InVolume, $OutVolume);
	my (@DailyStats) = "";
	my ($NumDays) = 0;

	print STDERR "INDEX=[$IndexFile]\n" if $Debug;
	print STDERR "TITLE=[$Title]\n" if $Debug;

	open (DIR, "cd $HtmlDir; find . -name \"[0-9]*.html\" -print | sort -r |") || 
		die "Error: $HtmlDir - $!\n";

	open (FILE, ">$IndexFile") || die "Error: $IndexFile - $!\n";

	print FILE <<EOT
<html>
<head>
<title>$Title1</title>
</head>
<body $BgColor>
<table>
<tr>
<td><a href=$WwwUrl> <img alt=$ImgText $ImgSize src=$ImgLogo></a></td>
<td><h3>$Title2</h3></td>
</tr>
</table>
EOT
;
		&PrintOutCopyright;
	print FILE <<EOT
<hr>
<p>
<table border=2>
<tr>
<th rowspan=2 colspan=2>Daily Statistics<th colspan=5>Incoming Feeds<th colspan=4>Outgoing Feeds
<tr>
<th>Accepted<th>Volume<th>Spam<th>Rejected<th>Errs<th>Accepted<th>Volume<th>Duplicates<th>Rejected
EOT
;

	while (<DIR>) {
		chop;
		s/^\.\///;
		if (/^(......)\.html/o && ($MaxDays == 0 || ($NumDays < $MaxDays))) {
			$NumDays++;
			$Date = $1;
			$FancyDate = &GetFancyNameDate ($Date, 1, 0);
			@DailyStats = &GetDailyStats ($Date);
			$InVolume = &GetVolume ($DailyStats[3]);
			$OutVolume = &GetVolume ($DailyStats[7]);
			print FILE <<EOT
<tr>
<td align="right"><strong>$NumDays</strong>
<td><a href="$Date.html"><tt>$FancyDate</tt></a>
<td align="right"><strong>$DailyStats[2]</strong>
<td align="right"><strong>$InVolume</strong>
<td align="right">$DailyStats[8]
<td align="right">$DailyStats[1]
<td align="right">$DailyStats[0]
<td align="right"><strong>$DailyStats[6]</strong>
<td align="right"><strong>$OutVolume</strong>
<td align="right">$DailyStats[4]
<td align="right">$DailyStats[5]
EOT
;
		}
	}
	print FILE <<EOT
</table>
<p>
<hr>
EOT
;
	&PrintOutFooter;

	close (FILE);
	close (DIR);
}

	
sub PrintSeparator
{
	my ($Num) = @_;

	if ($Num =~ /beg/i) {
		print FILE "<p>\n<hr>\n<p>\n";
	} elsif ($Num =~ /end/i) {
		print FILE "<a href=\"#00\">Goto top of page</a>, Goto top of table\n";
		print FILE "<p>\n<hr>\n";
	} else {
		print FILE "<a name=\"$Num\">\n";
		print FILE "<a href=\"#00\">Goto top of page</a>, Goto top of table\n";
		print FILE "<p>\n<hr>\n<p>\n";
	}
}


sub PrintIncomingFeeds
{
	my ($Title, $ListByVolume) = @_;
	my ($Num, $InCons, $InSecs, $InIhave, $InChk, $InErr, $InSpam, $InRej);
	my ($InRec, $InAdded, $InBytes, $InPercent, $InSubPercent, $Key, $Comment);
	my ($InPercentAcc, $InSubPercentAcc, $InPercentVol, $InSubPercentVol);
	my ($Field1, $Field2, $Field3, $Field4, $Total1, $Total2, $Total3, $Total4);
	my ($Format, $SubFormat, $AscFormat, $SubTotal1, $SubTotal2, $SubTotal3, $SubTotal4);

	&PrintTableHeader ($Title);

	if ($ListByVolume) {
		%OrderedList = %OrderByInBytes;
		$Field1 = "Volume"; 
		$Field2 = "%Vol";
		$Field3 = "Accepted"; 
		$Field4 = "%Acc";
		$AscFormat = "%8s %3.2f %8d %3.2f";
		$SubFormat = "%s%s %s%3.2f %s%d %s%3.2f";
		$Format = "%s%s%s%s %s%s%3.2f%s %s%s%d%s %s%s%3.2f%s";
	} else {
		%OrderedList = %OrderByInAdded;
		$Field1 = "Accepted"; 
		$Field2 = "%Acc";
		$Field3 = "Volume"; 
		$Field4 = "%Vol";
		$AscFormat = "%8d %3.2f %8s %3.2f";
		$SubFormat = "%s%d %s%3.2f %s%s %s%3.2f";
		$Format = "%s%s%d%s %s%s%3.2f%s %s%s%s%s %s%s%3.2f%s";
	}

	if ($TextFormat) {
		printf FILE "\n%-30s %8s %6s %8s %6s %8s %8s %8s %8s %8s %6s %6s\n\n",
			"Incoming Feed", $Field1, $Field2, $Field3, $Field4, "Check", "Ihave", "Spam", "Rejs", "Errs", "Cons", "HH:MM";
	} else {
		print FILE "<th colspan=2>Incoming Feed<th>$Field1<th>$Field2<th>$Field3<th>$Field4<th>Check<th>Ihave<th>Spam<th>Rejs<th>Errs<th>Cons<th>HH:MM\n";
	}

	$InAdded = $InBytes = 0;

	foreach $Key (reverse sort keys %OrderedList) {
		$Host = $OrderedList{$Key};

		if ($Feeds{$Host}->{InSecs} && $Feeds{$Host}->{InAdded}) {

			$InCons += $Feeds{$Host}->{InCons};
			$InSecs += $Feeds{$Host}->{InSecs};
			$InIhave += $Feeds{$Host}->{InIhave};
			$InChk += $Feeds{$Host}->{InChk};
			$InSpam += $Feeds{$Host}->{InSpam};
			$InErr += $Feeds{$Host}->{InErr};
			$InRej += $Feeds{$Host}->{InRej};
			$InAdded += $Feeds{$Host}->{InAdded};
			$InBytes += $Feeds{$Host}->{InBytes};
		}
	}

	if ($ListByVolume) {
		$Total1 = &GetVolume ($InBytes);
		$Total3 = $InAdded;
	} else {
		$Total1 = $InAdded;
		$Total3 = &GetVolume ($InBytes);
	}

	foreach $Key (reverse sort keys %OrderedList) {
		$Host = $OrderedList{$Key};

		if ($Feeds{$Host}->{InSecs} && $Feeds{$Host}->{InAdded}) {

			$InSubPercentAcc = &GetPercent ($InAdded, $Feeds{$Host}->{InAdded});
			$InPercentAcc += $InSubPercentAcc;

			$InSubPercentVol = &GetPercent ($InBytes, $Feeds{$Host}->{InBytes});
			$InPercentVol += $InSubPercentVol;

			if ($ListByVolume) {
				$SubTotal1 = &GetVolume ($Feeds{$Host}->{InBytes});
				$SubTotal2 = $InSubPercentVol;
				$SubTotal3 = $Feeds{$Host}->{InAdded};
				$SubTotal4 = $InSubPercentAcc;
			} else {
				$SubTotal1 = $Feeds{$Host}->{InAdded};
				$SubTotal2 = $InSubPercentAcc;
				$SubTotal3 = &GetVolume ($Feeds{$Host}->{InBytes});
				$SubTotal4 = $InSubPercentVol;
			}

			if ($TextFormat) {
				printf FILE "%-30s $AscFormat %8d %8d %8d %8d %8d %6d %6s\n",
					&GetHostLink ($Host), 
					$SubTotal1, 
					$SubTotal2, 
					$SubTotal3, 
					$SubTotal4,
					$Feeds{$Host}->{InChk},
					$Feeds{$Host}->{InIhave}, 
					$Feeds{$Host}->{InSpam}, 
					$Feeds{$Host}->{InRej}, 
					$Feeds{$Host}->{InErr}, 
					$Feeds{$Host}->{InCons}, 
					&GetHhMmSs ($Feeds{$Host}->{InSecs});
			} else {
				printf FILE "<tr>\n%s%s%d%s %s%s $SubFormat %s%d %s%d %s%d %s%d %s%d %s%d %s%s\n",
					$TDR, "<strong>", ++$Num, "</strong>",
					$TDL, &GetHostLink ($Host), 
					$TDR, $SubTotal1,
					$TDR, $SubTotal2,
					$TDR, $SubTotal3,
					$TDR, $SubTotal4,
					$TDR, $Feeds{$Host}->{InChk}, 
					$TDR, $Feeds{$Host}->{InIhave}, 
					$TDR, $Feeds{$Host}->{InSpam}, 
					$TDR, $Feeds{$Host}->{InRej},
					$TDR, $Feeds{$Host}->{InErr}, 
					$TDR, $Feeds{$Host}->{InCons}, 
					$TDR, &GetHhMmSs ($Feeds{$Host}->{InSecs});
			}
		}
	}

	if ($ListByVolume) {
		$Total2 = $InPercentVol;
		$Total4 = $InPercentAcc;
	} else {
		$Total2 = $InPercentAcc;
		$Total4 = $InPercentVol;
	}

	if ($TextFormat) {
		printf FILE "\n%s $AscFormat %8d %8d %8d %8d %8d %6d %6s\n\n",
			"                              ",
			$Total1, $Total2, $Total3, $Total4, $InChk, $InIhave, 
			$InSpam, $InRej, $InErr, $InCons, &GetHhMmSs ($InSecs), 
	} else {
		printf FILE "<tr>\n%s%s $Format %s%s%d%s %s%s%d%s %s%s%d%s %s%s%d%s %s%s%d%s %s%s%d%s %s%s%s%s\n",
			"<td align=\"center\" colspan=2>", "<strong>Total</strong>",
			$TDR, "<strong>", $Total1, "</strong>",
			$TDR, "<strong>", $Total2, "</strong>",
			$TDR, "<strong>", $Total3, "</strong>",
			$TDR, "<strong>", $Total4, "</strong>",
			$TDR, "<strong>", $InChk, "</strong>", 
			$TDR, "<strong>", $InIhave, "</strong>", 
			$TDR, "<strong>", $InSpam, "</strong>", 
			$TDR, "<strong>", $InRej, "</strong>", 
			$TDR, "<strong>", $InErr, "</strong>", 
			$TDR, "<strong>", $InCons, "</strong>", 
			$TDR, "<strong>", &GetHhMmSs ($InSecs), "</strong>";
	}

	$Comment = "INCOMING ERR=$InErr REJ=$InRej SPAM=$InSpam ADD=$InAdded VOL=$InBytes";

	&PrintTableFooter ($Comment);
}


sub PrintIncomingTimes
{
	my ($Title) = @_;
	my ($Num, $Hour, $InVolume, $InSubVolume);
	my ($InPercentAcc, $InSubPercentAcc, $InPercentVol, $InSubPercentVol);

	&PrintTableHeader ($Title);

	if ($TextFormat) {
		printf FILE "\n%5s %8s %6s %8s %6s %8s %8s %8s %8s %8s\n\n",
			"Hour", "Accepted", "%Acc", "Volume", "%Vol", "Check", "Ihave", "Spam", "Rejs", "Errs";
	} else {
		print FILE "<th>Hour<th>Accepted<th>%Acc<th>Volume<th>%Vol<th>Check<th>Ihave<th>Spam<th>Rejs<th>Errs\n";
	}

	&CalcHourlyStats;

	for ($Num = 0; $Num < 24; $Num++) {
		$Hour = sprintf ("%02d", $Num);

		$InSubPercentAcc = &GetPercent ($HourlyStats{Total}->{InAdded}, $HourlyStats{$Hour}->{InAdded});
		$InPercentAcc += $InSubPercentAcc;

		$InSubPercentVol = &GetPercent ($HourlyStats{Total}->{InBytes}, $HourlyStats{$Hour}->{InBytes});
		$InPercentVol += $InSubPercentVol;

		$InSubVolume = &GetVolume ($HourlyStats{$Hour}->{InBytes});

		if ($TextFormat) {
			printf FILE "%5s %8d %3.2f %8s %3.2f %8d %8d %8d %8d %8d\n",
				$Hour, 
				$HourlyStats{$Hour}->{InAdded}, 
				$InSubPercentAcc,
				$InSubVolume,
				$InSubPercentVol,
				$HourlyStats{$Hour}->{InChk},
				$HourlyStats{$Hour}->{InIhave},
				$HourlyStats{$Hour}->{InSpam},
				$HourlyStats{$Hour}->{InRej}, 
				$HourlyStats{$Hour}->{InErr};
		} else {
			printf FILE "<tr>\n%s%s%s%s %s%d %s%3.2f %s%s %s%3.2f %s%d %s%d %s%d %s%d %s%d\n",
				$TDR, "<strong>", $Hour, "</strong>",
				$TDR, $HourlyStats{$Hour}->{InAdded},
				$TDR, $InSubPercentAcc,
				$TDR, $InSubVolume,
				$TDR, $InSubPercentVol,
				$TDR, $HourlyStats{$Hour}->{InChk},
				$TDR, $HourlyStats{$Hour}->{InIhave},
				$TDR, $HourlyStats{$Hour}->{InSpam},
				$TDR, $HourlyStats{$Hour}->{InRej},
				$TDR, $HourlyStats{$Hour}->{InErr};
		}
	}

	$InVolume = &GetVolume ($HourlyStats{Total}->{InBytes});

	if ($TextFormat) {
		printf FILE "\n%5s %8d %3.2f %8s %3.2f %8d %8d %8d %8d %8d\n\n",
			"Total",
			$HourlyStats{Total}->{InAdded},
			$InPercentAcc,
			$InVolume,
			$InPercentVol,
			$HourlyStats{Total}->{InChk},
			$HourlyStats{Total}->{InIhave},
			$HourlyStats{Total}->{InSpam},
			$HourlyStats{Total}->{InRej},
			$HourlyStats{Total}->{InErr};
	} else {
		printf FILE "<tr>\n%s%s%s%s %s%s%d%s %s%s%3.2f%s %s%s%s%s %s%s%3.2f%s %s%s%d%s %s%s%d%s %s%s%d%s %s%s%d%s %s%s%d%s\n",
			$TDR, "<strong>", "Total", "</strong>",
			$TDR, "<strong>", $HourlyStats{Total}->{InAdded}, "</strong>",
			$TDR, "<strong>", $InPercentAcc, "</strong>",
			$TDR, "<strong>", $InVolume, "</strong>",
			$TDR, "<strong>", $InPercentVol, "</strong>",
			$TDR, "<strong>", $HourlyStats{Total}->{InChk}, "</strong>",
			$TDR, "<strong>", $HourlyStats{Total}->{InIhave}, "</strong>",
			$TDR, "<strong>", $HourlyStats{Total}->{InSpam}, "</strong>",
			$TDR, "<strong>", $HourlyStats{Total}->{InRej}, "</strong>",
			$TDR, "<strong>", $HourlyStats{Total}->{InErr}, "</strong>";
	}

	&PrintTableFooter ("");
}


sub PrintOutgoingFeeds
{
	my ($Title, $ListByVolume) = @_;
	my ($Key, $Comment, $OutCons, $OutSecs, $OutDup, $OutRej, $OutAcc, $OutBytes);
	my ($Num, $OutPercentAcc, $OutSubPercentAcc, $OutPercentVol, $OutSubPercentVol); 
	my ($Field1, $Field2, $Field3, $Field4, $Total1, $Total2, $Total3, $Total4);
	my ($Format, $SubFormat, $SubTotal1, $SubTotal2, $SubTotal3, $SubTotal4);

	&PrintTableHeader ($Title);

	if ($ListByVolume) {
		%OrderedList = %OrderByOutBytes;
		$Field1 = "Volume"; 
		$Field2 = "%Vol";
		$Field3 = "Accepted"; 
		$Field4 = "%Total";
		$SubFormat = "%s%s %s%3.2f %s%d %s%3.2f";
		$Format = "%s%s%s%s %s%s%3.2f%s %s%s%d%s %s%s%3.2f%s";
	} else {
		%OrderedList = %OrderByOutAcc;
		$Field1 = "Accepted"; 
		$Field2 = "%Total";
		$Field3 = "Volume"; 
		$Field4 = "%Vol";
		$SubFormat = "%s%d %s%3.2f %s%s %s%3.2f";
		$Format = "%s%s%d%s %s%s%3.2f%s %s%s%s%s %s%s%3.2f%s";
	}

	if ($TextFormat) {
		printf FILE "\n%-30s %8s %6s %9s %8s %8s %5s %8s %5s\n\n",
			"Outgoing Feed", "Cons", "HH:MM", 
			"Duplicates", "Rejs", "Accepted", 
			"%Total", "Volume", "%Vol";
	} else {
		print FILE "<th colspan=2>Outgoing Feed<th>$Field1<th>$Field2<th>$Field3<th>$Field4<th>Duplicates<th>Rejs<th>Cons<th>HH:MM\n";
	}

	$OutAcc = $OutBytes = 0;

	foreach $Key (reverse sort keys %OrderedList) {
		$Host = $OrderedList{$Key};

		if ($Feeds{$Host}->{OutSecs} && $Feeds{$Host}->{OutAcc}) {

			$OutCons += $Feeds{$Host}->{OutCons};
			$OutSecs += $Feeds{$Host}->{OutSecs};
			$OutDup += $Feeds{$Host}->{OutDup};
			$OutRej += $Feeds{$Host}->{OutRej};
			$OutAcc += $Feeds{$Host}->{OutAcc};
			$OutBytes += $Feeds{$Host}->{OutBytes};
		}
	}

	if ($ListByVolume) {
		$Total1 = &GetVolume ($OutBytes);
		$Total3 = $OutAcc;
	} else {
		$Total1 = $OutAcc;
		$Total3 = &GetVolume ($OutBytes);
	}

	foreach $Key (reverse sort keys %OrderedList) {
		$Host = $OrderedList{$Key};

		if ($Feeds{$Host}->{OutSecs} && $Feeds{$Host}->{OutAcc}) {

			$OutSubPercentAcc = &GetPercent ($OutAcc, $Feeds{$Host}->{OutAcc});
			$OutPercentAcc += $OutSubPercentAcc;

			$OutSubPercentVol = &GetPercent ($OutBytes, $Feeds{$Host}->{OutBytes});
			$OutPercentVol += $OutSubPercentVol;

			if ($ListByVolume) {
				$SubTotal1 = &GetVolume ($Feeds{$Host}->{OutBytes});
				$SubTotal2 = $OutSubPercentVol;
				$SubTotal3 = $Feeds{$Host}->{OutAcc};
				$SubTotal4 = $OutSubPercentAcc;
			} else {
				$SubTotal1 = $Feeds{$Host}->{OutAcc};
				$SubTotal2 = $OutSubPercentAcc;
				$SubTotal3 = &GetVolume ($Feeds{$Host}->{OutBytes});
				$SubTotal4 = $OutSubPercentVol;
			}

			if ($TextFormat) {
				printf FILE "%-30s %8d %6s %9d %8d %8d %3.2f %8s %3.2f\n",
					&GetHostLink ($Host),  
					$Feeds{$Host}->{OutCons}, 
					&GetHhMmSs ($Feeds{$Host}->{OutSecs}),
					$Feeds{$Host}->{OutDup}, 
					$Feeds{$Host}->{OutRej}, 
					$Feeds{$Host}->{OutAcc}, 
					$OutSubPercentAcc,
					&GetVolume ($Feeds{$Host}->{OutBytes}), 
					$OutSubPercentVol;
			} else {
				printf FILE "<tr>\n%s%s%d%s %s%s $SubFormat %s%d %s%d %s%d %s%s\n",
					$TDR, "<strong>", ++$Num, "</strong>",
					$TDL, &GetHostLink ($Host),
					$TDR, $SubTotal1,
					$TDR, $SubTotal2,
					$TDR, $SubTotal3,
					$TDR, $SubTotal4,
					$TDR, $Feeds{$Host}->{OutDup}, 
					$TDR, $Feeds{$Host}->{OutRej}, 
					$TDR, $Feeds{$Host}->{OutCons}, 
					$TDR, &GetHhMmSs ($Feeds{$Host}->{OutSecs});
			}
		}
	}

	if ($ListByVolume) {
		$Total2 = $OutPercentVol;
		$Total4 = $OutPercentAcc;
	} else {
		$Total2 = $OutPercentAcc;
		$Total4 = $OutPercentVol;
	}

	if ($TextFormat) {
		printf FILE "\n%s %8d %6s %9d %8d %8d %3.2f %8s %3.2f\n\n",
			"                              ",
			$OutCons, &GetHhMmSs ($OutSecs), $OutDup, $OutRej, $OutAcc, 
			$OutPercentAcc, &GetVolume ($OutBytes), $OutPercentVol;
	} else {
		printf FILE "<tr>\n%s%s $Format %s%s%d%s %s%s%d%s %s%s%d%s %s%s%s%s\n",
			"<td align=\"center\" colspan=2>", "<strong>Total</strong>",
			$TDR, "<strong>", $Total1, "</strong>",
			$TDR, "<strong>", $Total2, "</strong>",
			$TDR, "<strong>", $Total3, "</strong>",
			$TDR, "<strong>", $Total4, "</strong>",
			$TDR, "<strong>", $OutDup, "</strong>",
			$TDR, "<strong>", $OutRej, "</strong>",
			$TDR, "<strong>", $OutCons, "</strong>",
			$TDR, "<strong>", &GetHhMmSs ($OutSecs), "</strong>";
	}

	$Comment = "OUTGOING DUP=$OutDup REJ=$OutRej ACC=$OutAcc VOL=$OutBytes";

	&PrintTableFooter ($Comment);
}


sub PrintQueueStats
{
	my ($Title) = @_;
	my ($OutCons, $OutSecs, $OutDup, $OutRej, $OutAcc);
	my ($Key, $Color, $TDL, $TDR);

	if (! $TextFormat) {
		print FILE <<EOT
<table border=2 width=100%>
<td align="center" bgcolor=$QueColor1><strong>&gt;= $QuePercent1% Full</strong></td>
<td align="center" bgcolor=$QueColor2><strong>&gt;= $QuePercent2% Full</strong></td>
<td align="center" bgcolor=$QueColor3><strong>&gt;= $QuePercent3% Full</strong></td>
<td align="center" bgcolor=$QueColor4><strong>&gt;= $QuePercent4% Full</strong></td>
<td align="center" bgcolor=$QueColor5><strong>&gt;= $QuePercent5% Full</strong></td>
<td align="center" bgcolor=$QueColor6><strong>&gt;= $QuePercent6% Full</strong></td>
</table>
<p>
EOT
;
	}

	&PrintTableHeader ($Title);

	if ($TextFormat) {
		printf FILE "\n%-30s %19s %9s %9s %9s\n\n",
			"Outgoing Feed", "Batch Seq", 
			"Batch Num", "Batch Max", "%Full";
	} else {
		print FILE "<th>Outgoing Feed<th>Batch Seq<th>Batch Num<th>Batch Max<th>%Full\n";
	}

	foreach $Key (reverse sort keys %OrderByBatchFull) {
		$Host = $OrderByBatchFull{$Key};

		if ($Feeds{$Host}->{BatchMax}) {

			if ($TextFormat) {
				printf FILE "%-30s %9d-%9d %9d %9d %9d\n",
				        &GetQueueHost ($Host),
					$Feeds{$Host}->{BatchBeg}, 
					$Feeds{$Host}->{BatchEnd},
					$Feeds{$Host}->{BatchNum},
					$Feeds{$Host}->{BatchMax},
					$Feeds{$Host}->{BatchFull};
			} else {
				if ($Feeds{$Host}->{BatchFull} >= $QuePercent6) {
					$Color = "bgcolor=$QueColor6";
				} elsif ($Feeds{$Host}->{BatchFull} >= $QuePercent5) {
					$Color = "bgcolor=$QueColor5";
				} elsif ($Feeds{$Host}->{BatchFull} >= $QuePercent4) {
					$Color = "bgcolor=$QueColor4";
				} elsif ($Feeds{$Host}->{BatchFull} >= $QuePercent3) {
					$Color = "bgcolor=$QueColor3";
				} elsif ($Feeds{$Host}->{BatchFull} >= $QuePercent2) {
					$Color = "bgcolor=$QueColor2";
				} elsif ($Feeds{$Host}->{BatchFull} >= $QuePercent1) {
					$Color = "bgcolor=$QueColor1";
				} else {
					$Color = "";
				}
				$TDL = "<td $Color>";
				$TDR = "<td $Color align=\"right\">";

				printf FILE "<tr>\n%s%s %s%d-%d %s%d %s%d %s%d\n",
				        $TDL, &GetQueueHost ($Host),
					$TDR, $Feeds{$Host}->{BatchBeg}, $Feeds{$Host}->{BatchEnd},
					$TDR, $Feeds{$Host}->{BatchNum},
					$TDR, $Feeds{$Host}->{BatchMax},
					$TDR, $Feeds{$Host}->{BatchFull};
			}
		}
	}

	&PrintTableFooter ("");
}


sub PrintCuriousActivity
{
	my ($Title) = @_;
	my ($Key, $BegTime, $EndTime);

	&PrintTableHeader ($Title);

	if ($TextFormat) {
		printf FILE "\n%s %s %-8s %50s\n\n",
			"1st Time", "Last Time", "# Msgs", "Message";
	} else {
		print FILE "<th>1st Time<th>Last Time<th># Msgs<th>Message\n";
	}

	foreach $Key (reverse sort keys %OrderByErrorTotal) {
		$Msg = $OrderByErrorTotal{$Key};
		
		if ($ErrorList{$Msg}->{Tot}) {

			$BegTime = $ErrorList{$Msg}->{BegDate};
			$EndTime = $ErrorList{$Msg}->{EndDate};

			if ($TextFormat) {
				printf FILE "%s %s %8d %-50s\n",
					$BegTime, $EndTime, $ErrorList{$Msg}->{Tot}, $Msg;
			} else {
				printf FILE "<tr>\n%s%s %s%s %s%d %s%s\n",
					$TDL, $BegTime,
					$TDL, $EndTime,
					$TDR, $ErrorList{$Msg}->{Tot},
					$TDL, $Msg;
			}
		}
	}

	&PrintTableFooter ("");
}


sub PrintSpamActivity
{
	my ($Title) = @_;
	my ($Key, $Time, $Num, $Host, $ByDups, $ByRate, $Total);

	&PrintTableHeader ($Title);

	if ($TextFormat) {
		printf FILE "\n%-50s %8d %8d %8d\n\n",
			"Host", "By Post Rate", "By Dup Body", "# Articles";
	} else {
		print FILE "<th>Host<th>By Post Rate<th>By Dup Body<th># Articles\n";
	}

	$Num = 0;
	foreach $Key (reverse sort keys %OrderBySpamTotal) {
		if ($Num < $MaxSpamList) {

			$Host = $OrderBySpamTotal{$Key};
	
			$Total = $SpamList{$Host}->{Total};
			$ByRate = $SpamList{$Host}->{ByRate};
			$ByDups = $SpamList{$Host}->{ByDups};
#			print "SPAM=[$Key] $Host\n";

			if ($Total) {
				if ($TextFormat) {
					printf FILE "%-50s %8d %8d %8d\n",
						$Host, $ByRate, $ByDups, $Total; 
				} else {	
					printf FILE "<tr>\n%s%s %s%d %s%d %s%d\n",
						$TDL, $Host,
						$TDR, $ByRate,
						$TDR, $ByDups,
						$TDR, $Total;
				}
			}
			$Num++;
		} else {
			goto SpamDone;
		}
	}
SpamDone:

	&PrintTableFooter ("");
}


sub PrintTableHeader
{
	my ($Title) = @_;

	if ($TextFormat) {
		print FILE "$Title\n";
	} else {
		print FILE <<EOT
<table border=2 width=100%>
<caption><h3>$Title</h3></caption>
EOT
;
	}
}


sub PrintTableFooter
{
	my ($Comment) = @_;

	if (! $TextFormat) {
		print FILE "</table>\n";
		print FILE "<!-- $Comment -->\n" if $Comment ne "";
	}
}


sub PrintOutHeader
{
	my ($TimePeriod) = $BegDate;
	my ($Title, $Title2);

	if ($BegDate ne $EndDate) {
		$TimePeriod = "$BegDate - $EndDate";
	} else {
		$TimePeriod = &GetFancyNameDate ($BegDate, 1, 0);
	}

	$Title = "DIABLO news statistics for '$DiabloHost' for $TimePeriod";
	$Title2 = "<a href=$DiabloUrl>DIABLO</a>  news statistics for '$DiabloHost' for $TimePeriod";

	if ($TextFormat) {
		print FILE "$Title\n\n\n";
	} else {
		print FILE <<EOT
<html>
<head>
<title>$Title</title>
</head>
<body $BgColor>
<a name="00">
<table>
<tr>
<td><a href=$WwwUrl> <img alt=$ImgText $ImgSize src=$ImgLogo></a></td>
<td><h3>$Title2</h3></td>
</tr>
</table>
EOT
;
		&PrintOutCopyright;
		print FILE <<EOT
<hr>
<p>
<ol>
<li><a href="#01">Summary of Incoming Feeds by Article</a>
<li><a href="#02">Summary of Incoming Feeds by Volume</a>
<li><a href="#03">Summary of Incoming Feeds by Time</a>
<li><a href="#04">Summary of Outgoing Feeds by Article</a>
<li><a href="#05">Summary of Outgoing Feeds by Volume</a>
<li><a href="#06">Summary of Outgoing Queue</a>
<li><a href="#07">Summary of Curious Activity</a>
<li><a href="#08">Summary of SPAM Activity (Top $MaxSpamList of $TotalSpamList)</a>
</ol>
EOT
;
	}
}


sub PrintOutFooter
{
	&PrintOutCopyright;

	if (! $TextFormat) {
		print FILE "</body>\n</html>\n";
	}
}


sub PrintOutCopyright
{
	if ($TextFormat) {
		print FILE "\nGenerated on $GenDate $GenTime by $ScriptName $Version. Copyright (C) 1997 Iain Lea (iain\@ecrc.de)\n";
	} else {
		print FILE <<EOT
Generated on $GenDate $GenTime by 
<a href="$FtpUrl">$ScriptName $Version</a>. 
Copyright &copy; 1997 
<a href="mailto:iain\@ecrc.de">Iain Lea</a>. 
EOT
;
	}
}


sub GetYyMmDd
{
	my ($Date) = @_;

	# Apr 10 11:36:54
	if ($Date =~ /^(...) (..)/) {
		return sprintf "%02d%02d%02d", $GenYear, $MonthNumByName{$1}, $2;
	} else {
		return "";
	}
}


sub GetHhMmSs
{
	my ($Secs) = @_;
#	my ($Time) = "00:00:00";
	my ($Time) = "00:00";
	my ($Hh, $Mm, $Ss);

	if ($Secs) {
		$Hh = int($Secs / 3600);
		$Mm = int(($Secs - ($Hh * 3600)) / 60);
		$Ss = $Secs % 60;

#		$Time = sprintf "%02d:%02d:%02d", $Hh, $Mm, $Ss;
		$Time = sprintf "%02d:%02d", $Hh, $Mm;
	}

#	print "Secs=[$Secs] Time=[$Time]\n";

	return $Time;
}


sub GetFancyDate
{
	my ($Date) = @_;

	if (/^(..)(..)(..)/) {
		$Date = sprintf "%s %s %s", $3, $MonthNameByNum{$2}, $1;
	}

	return $Date;
}


# $Date = &GetFancyNameDate ("960109", [0|1], [0|1]);
# par2: 0 = Long / 1 = Short form of dayname ie. Monday / Mon
# par3: 0 = not split / 1 = split returned line ie. Mon\n14 Oct 95

sub GetFancyNameDate 
{
	local ($CurDate, $ShortName, $SplitLine) = @_;
	local ($Break, $DayName);

	if ($SplitLine) { 
		$Break = '<br>'; 
	} else { 
		$Break = ' '; 
	}
	$CurDate =~ /^(..)(..)(..)/;
	$DayName = &GetDayOfWeek ($1, $2, $3, $ShortName);

	return "$DayName$Break$3 $MonthNameByNum{$2} $1";
}


sub GetDayOfWeek 
{
	# Parameters: YY, MM, DD, [0|1]
	local($Yy, $Mm, $Dd, $ShortName) = @_;
	local($DayNum);
    
	$DayNum = (localtime(timelocal(0,0,0,$Dd,$Mm-1,$Yy,0,0)))[6];

	if ($ShortName) { 
		return $DayNameShort[$DayNum]; 
	} else { 
		return $DayName[$DayNum]; 
	}
}


sub GetDailyStats
{
	my ($Date) = @_;
	my ($InErr, $InRej, $InSpam, $InTot, $InVol, $OutDup, $OutRej, $OutAcc, $OutVol);

	$InErr = $InRej = $InSpam = $InTot = $InVol = $OutDup = $OutRej = $OutAcc = $OutVol = 0;

	if (open (STATS, "$Date.html")) {
		while (<STATS>) {
			if (/INCOMING ERR=([\ 0-9]*) REJ=([\ 0-9]*) SPAM=([\ 0-9]*) ADD=([\ 0-9]*) VOL=([\ 0-9]*)/) {
				$InErr = $1;
				$InRej = $2;
				$InSpam = $3;
				$InTot = $4;
				$InVol = $5;
			} elsif (/INCOMING ERR=([\ 0-9]*) REJ=([\ 0-9]*) ADD=([\ 0-9]*) VOL=([\ 0-9]*)/) {
				$InErr = $1;
				$InRej = $2;
				$InTot = $3;
				$InVol = $4;
			} elsif (/INCOMING ERR=([\ 0-9]*) REJ=([\ 0-9]*) ADD=([\ 0-9]*)/) {	# < v1.10
				$InErr = $1;
				$InRej = $2;
				$InTot = $3;
			}
			if (/OUTGOING DUP=([\ 0-9]*) REJ=([\ 0-9]*) ACC=([\ 0-9]*) VOL=([\ 0-9]*)/) {
				$OutDup = $1;
				$OutRej = $2;
				$OutAcc = $3;
				$OutVol = $4;
			} elsif (/OUTGOING DUP=([\ 0-9]*) REJ=([\ 0-9]*) ACC=([\ 0-9]*)/) {	# < v1.10
				$OutDup = $1;
				$OutRej = $2;
				$OutAcc = $3;
			}
		}
		close STATS;
	} else {
		warn "Error: $Date.html - $!\n";
	}
	
	return ($InErr, $InRej, $InTot, $InVol, $OutDup, $OutRej, $OutAcc, $OutVol, $InSpam);
}


sub GetPercent
{
	my ($Total, $SubTotal) = @_;

	return ($SubTotal && $Total) ? ($SubTotal / $Total) * 100 : 0; 	
}


sub AddToInTimeList
{
	my ($Date, $Host, $InAdded, $InBytes, $InChk, $InIhave, $InSpam, $InRej, $InErr) = @_;

# print "IN:  $Host, $InAdded, $InBytes, $InChk, $InIhave, $InSpam, $InRej, $InErr\n";

	# ... 23:59:59
	if ($Date =~ / (..):..:../) {
		$Hour = $1;
		$Feeds{$Host}->{$Hour}->{InAdded} += $InAdded;
		$Feeds{$Host}->{$Hour}->{InBytes} += $InBytes;
		$Feeds{$Host}->{$Hour}->{InChk} += $InChk;
		$Feeds{$Host}->{$Hour}->{InIhave} += $InIhave;
		$Feeds{$Host}->{$Hour}->{InSpam} += $InSpam;
		$Feeds{$Host}->{$Hour}->{InRej} += $InRej;
		$Feeds{$Host}->{$Hour}->{InErr} += $InErr;
# print "IN: $Hour  $Host  $Feeds{$Host}->{$Hour}->{InAdded}  $Feeds{$Host}->{$Hour}->{InBytes}\n";
	}
}


sub AddToOutTimeList
{
	my ($Date, $Host, $OutAcc, $OutBytes, $OutDup, $OutRej) = @_;

	# ... 23:59:59
	if ($Date =~ / (..):..:../) {
		$Hour = $1;
		$Feeds{$Host}->{$Hour}->{OutAcc} += $OutAcc;
		$Feeds{$Host}->{$Hour}->{OutBytes} += $OutBytes;
		$Feeds{$Host}->{$Hour}->{OutDup} += $OutDup;
		$Feeds{$Host}->{$Hour}->{OutRej} += $OutRej;
	}
}


sub AddToErrorList
{
	my ($Error, $FullDate, $LineDate) = @_;

	$ErrorList{$Error}->{Tot} += 1;
	if ($ErrorList{$Error}->{BegDate} eq "") {
		$ErrorList{$Error}->{BegDate} = $FullDate;
	} else {
		$ErrorList{$Error}->{EndDate} = $FullDate;
	}

	print "$LineDate  $Error\n" if $Debug;
}


sub AddToSpamList
{
	my ($Spam, $FullDate, $LineDate) = @_;
	my ($Num, $Host, $MsgId, $ByDups, $ByRate, $ByFeed);

	$ByDups = $ByRate = $ByFeed = 0;

	# ... by-post-rate copy #33: <5rplh4$12p@ron.ipa.com> 206.1.57.29
	# ... by-dup-body copy #1: <5rpvin$8ro$1@telemedia.de> essn.mediaways.net
	# ... dnewsfeeds copy #-1: <5r4dpg$le8$1@news.utu.fi> apus.astro.utu.fi
	if ($Spam =~ /by-post-rate copy #([\-0-9]*):\s+([A-z0-9\.\-_\<>\$@]*)\s+([A-z0-9\.\-_]*)/) {
		$Num = $1;
		$MsgId = $2;
		$Host = $3;
		$ByRate = 1;
	} elsif ($Spam =~ /by-dup-body copy #([\-0-9]*):\s+([A-z0-9\.\-_\<>\$@]*)\s+([A-z0-9\.\-_]*)/) {
		$Num = $1;
		$MsgId = $2;
		$Host = $3;
		$ByDups = 1;
	} elsif ($Spam =~ /dnewsfeeds copy #([\-0-9]*):\s+([A-z0-9\.\-_\<>\$@]*)\s+([A-z0-9\.\-_]*)/) {
		$Num = $1;
		$MsgId = $2;
		$Host = $3;
		$ByFeed = 1;
	}
	if ($ByRate || $ByDups || $ByFeed) {
		$SpamList{$Host}->{Total} += 1;
		$SpamList{$Host}->{ByRate} += $ByRate;
		$SpamList{$Host}->{ByDups} += $ByDups;
		$SpamList{$Host}->{ByFeed} += $ByFeed;

		print "SPAM:  Type=$ByDups ($SpamList{$Host}->{ByDups})| $ByRate ($SpamList{$Host}->{ByRate})| $ByFeed ($SpamList{$Host}->{ByFeed})  MsgID=$MsgId  Site=$Host\n" if $Debug;
	}
}


sub GetVolume 
{
	# This is disgusting, but fcmp does lexical compares of floats expressed in
	# scientific notation, which is completely useless... If I knew more Perl I
	# could probably compute <X>Int from <X> instead of having duplicate defines.

#print "GetVolume(@_)\n";

	@_[0] += 0; # In case @_[0] is empty

	my ($Bytes) = new Math::BigFloat @_;
	my ($BytesInt) = new Math::BigInt @_;
	my ($Vol) = new Math::BigFloat "0.0";
	my ($Gb) = new Math::BigFloat "1073741824.0";
	my ($GbInt) = new Math::BigInt "1073741824";
	my ($Mb) = new Math::BigFloat "1048576.0";
	my ($MbInt) = new Math::BigInt "1048576";
	my ($Kb) = new Math::BigFloat "1024.0";
	my ($KbInt) = new Math::BigInt "1024";
	my ($IsKb, $IsMb, $IsGb);

#print "Bytes=[$Bytes]  BytesInt=[$BytesInt]";

	$IsKb = Math::BigInt::cmp ($BytesInt, $KbInt);
	$IsMb = Math::BigInt::cmp ($BytesInt, $MbInt);
	$IsGb = Math::BigInt::cmp ($BytesInt, $GbInt);

#print "\nCMP=$BytesInt  KB=$KbInt RET=[$IsKb]  MB=$MbInt RET=[$IsMb]  GB=$GbInt RET=[$IsGb]\n";

	if ($IsGb == 0 || $IsGb == 1) {
		$Sign = "GB";
		$Vol = $Bytes / $Gb;
	} elsif ($IsMb == 0 || $IsMb == 1) {
CalcMb:
		$Sign = "MB";
		$Vol = $Bytes / $Mb;
	} else {
CalcKb:
		$Sign = "KB";
		$Vol = $Bytes / $Kb;
	}

#printf "Num=$Bytes  Format(%g)=%3.2f%s\n", $Vol, $Vol, $Sign;

	$Vol = sprintf "%3.2f%s", $Vol, $Sign;

	# These should no longer be needed
	goto CalcMb if ($Vol =~ /^0\.[1-9][0-9]Gb/);
	goto CalcKb if ($Vol =~ /^0\.0[0-9]Gb/);

	return $Vol;
}


sub GetQueueHost
{
	my ($QHost) = @_;
	my ($QLine, $First, $Second, $Rest);

	$QHost =~ s/[ \t]+//;

	if ($QueueHost) {
		open (QFILE, $CtlFile) || die "Error: $CtlFile - $!\n";

		do {
			$QLine = <QFILE>;
			chop ($QLine);
			($First, $Second, $Rest) = split(/[ \t]+/, $QLine, 3);

		} until ($First eq $QHost || eof);

		close (QFILE);

		if ($Second eq '') {
			$Rest = $QHost;
		}
		$Rest = $Second.'  ('.$QHost.')';

		return $Rest;
	} else {
		return $QHost;
	}
}


sub MakeVolKey                                 
{
	my ($Bytes, $Num) = @_;
                
	until (length($Bytes) > 9) {                                        
		$Bytes = " " . $Bytes;
	}
	until (length($Num) > 2) {
		$Num = " " . $Num;                                   
	}
                
	return "$Bytes.$Num";                                                
}


sub GetHostLink
{
	my ($Host) = @_;
	my ($Link, $HostLink);

	$HostLink = $Host;
	if ($HostLink) {
		$Link = $HostLinkList{$Host};

		if ($Link) {
			$HostLink = sprintf ("<a href=\"%s\">%s</a>", $Link, $Host);
		}
		print "HOST:  Host=[$Host]  Link=[$Link]  HostLink=[$HostLink]\n" if $Debug;
	}

	return $HostLink;
}


sub ReadCfgFile
{
	if (open (FILE, $CfgFile)) {
		while (<FILE>) {
			if (/^link:\s+([A-z0-9_\.\/\-\~:]*)\s+([A-z0-9_\.\/\-\~:]*)\n/) {
				$Host = $1;
				$Link = $2;
				$HostLinkList{$Host} = $Link;
				print "LINK:  [$Host]==[$HostLinkList{$Host}]\n" if $Debug;
			}
		}
		close FILE;
	}
}


sub CalcHourlyStats
{
	my ($Num, $Hour);

	for ($Num = 0; $Num < 24; $Num++) {
		$Hour = sprintf ("%02d", $Num);

		foreach $Host (keys %Feeds) {
			$HourlyStats{$Hour}->{InAdded} += $Feeds{$Host}->{$Hour}->{InAdded};
			$HourlyStats{$Hour}->{InBytes} +=  $Feeds{$Host}->{$Hour}->{InBytes};
			$HourlyStats{$Hour}->{InChk} += $Feeds{$Host}->{$Hour}->{InChk};
			$HourlyStats{$Hour}->{InIhave} += $Feeds{$Host}->{$Hour}->{InIhave};
			$HourlyStats{$Hour}->{InSpam} += $Feeds{$Host}->{$Hour}->{InSpam};
			$HourlyStats{$Hour}->{InRej} += $Feeds{$Host}->{$Hour}->{InRej};
			$HourlyStats{$Hour}->{InErr} += $Feeds{$Host}->{$Hour}->{InErr};
# print "HOST: [$Hour]  [$Host]  $HourlyStats{$Hour}->{InAdded}  $Feeds{$Host}->{$Hour}->{InAdded}\n";
		}
		$HourlyStats{Total}->{InAdded} += $HourlyStats{$Hour}->{InAdded};
		$HourlyStats{Total}->{InBytes} += $HourlyStats{$Hour}->{InBytes};
		$HourlyStats{Total}->{InChk} += $HourlyStats{$Hour}->{InChk};
		$HourlyStats{Total}->{InIhave} += $HourlyStats{$Hour}->{InIhave};
		$HourlyStats{Total}->{InSpam} += $HourlyStats{$Hour}->{InSpam};
		$HourlyStats{Total}->{InRej} += $HourlyStats{$Hour}->{InRej};
		$HourlyStats{Total}->{InErr} += $HourlyStats{$Hour}->{InErr};

#		printf "HOUR: [$Hour]  Acc=%d Vol=%ld Chk=%d Ihave=%d Spam=%d Rejs=%d Errs=%d\n",
#			$HourlyStats{$Hour}->{InAdded}, $HourlyStats{$Hour}->{InBytes},
#			$HourlyStats{$Hour}->{InChk}, $HourlyStats{$Hour}->{InIhave},
#			$HourlyStats{$Hour}->{InSpam}, $HourlyStats{$Hour}->{InRej},
#			$HourlyStats{$Hour}->{InErr};
	}
#	printf "TOTAL:  Acc=%d Vol=%ld Chk=%d Ihave=%d Spam=%d Rejs=%d Errs=%d\n",
#		$HourlyStats{Total}->{InAdded}, $HourlyStats{Total}->{InBytes},
#		$HourlyStats{Total}->{InChk}, $HourlyStats{Total}->{InIhave},
#		$HourlyStats{Total}->{InSpam}, $HourlyStats{Total}->{InRej},
#		$HourlyStats{Total}->{InErr};
}


#########################

sub PrintIncomingGraph
{
	my ($Title, $ListByVolume) = @_;

	if ($ListByVolume) {
		%OrderedList = %OrderByInBytes;
	} else {
		%OrderedList = %OrderByInAdded;
	}

	foreach $Key (reverse sort keys %OrderedList) {
		$Host = $OrderedList{$Key};

		if ($Feeds{$Host}->{InSecs} && $Feeds{$Host}->{InAdded}) {

		}
	}
}



